@@ -1,6 +1,6 @@
 //! Genesis public interface and main loop
 
-use crate::audio::GenesisAudioResampler;
+use crate::audio::{GenesisAudioResampler, LowPassFilter};
 use crate::input::{GenesisInputs, InputState};
 use crate::memory::{Cartridge, MainBus, MainBusSignals, MainBusWrites, Memory};
 use crate::timing::{CycleCounters, GenesisCycleCounters};
@@ -13,7 +13,7 @@ use jgenesis_common::frontend::{
     SaveWriter, TickEffect, TimingMode,
 };
 use jgenesis_common::num::GetBit;
-use jgenesis_proc_macros::{EnumDisplay, EnumFromStr};
+use jgenesis_proc_macros::{EnumAll, EnumDisplay, EnumFromStr};
 use m68000_emu::M68000;
 use smsgg_core::psg::{Sn76489, Sn76489TickEffect, Sn76489Version};
 use std::cmp;
@@ -34,8 +34,11 @@ pub enum GenesisError<RErr, AErr, SErr> {
 
 pub type GenesisResult<RErr, AErr, SErr> = Result<TickEffect, GenesisError<RErr, AErr, SErr>>;
 
-#[derive(Debug, Clone, Copy, PartialEq, Eq, Default, Encode, Decode, EnumDisplay, EnumFromStr)]
+#[derive(
+    Debug, Clone, Copy, PartialEq, Eq, Default, Encode, Decode, EnumDisplay, EnumFromStr, EnumAll,
+)]
 #[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]
+#[cfg_attr(feature = "clap", derive(jgenesis_proc_macros::CustomValueEnum))]
 pub enum GenesisAspectRatio {
     #[default]
     Ntsc,
@@ -73,8 +76,9 @@ impl GenesisAspectRatio {
     }
 }
 
-#[derive(Debug, Clone, Copy, PartialEq, Eq, EnumDisplay, EnumFromStr, Encode, Decode)]
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Encode, Decode, EnumDisplay, EnumAll)]
 #[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]
+#[cfg_attr(feature = "clap", derive(jgenesis_proc_macros::CustomValueEnum))]
 pub enum GenesisRegion {
     Americas,
     Japan,
@@ -144,8 +148,14 @@ pub struct GenesisEmulatorConfig {
     pub emulate_non_linear_vdp_dac: bool,
     pub render_vertical_border: bool,
     pub render_horizontal_border: bool,
+    pub plane_a_enabled: bool,
+    pub plane_b_enabled: bool,
+    pub sprites_enabled: bool,
+    pub window_enabled: bool,
+    pub backdrop_enabled: bool,
     pub quantize_ym2612_output: bool,
     pub emulate_ym2612_ladder_effect: bool,
+    pub low_pass_filter: LowPassFilter,
     pub ym2612_enabled: bool,
     pub psg_enabled: bool,
 }
@@ -158,6 +168,11 @@ impl GenesisEmulatorConfig {
             emulate_non_linear_dac: self.emulate_non_linear_vdp_dac,
             render_vertical_border: self.render_vertical_border,
             render_horizontal_border: self.render_horizontal_border,
+            plane_a_enabled: self.plane_a_enabled,
+            plane_b_enabled: self.plane_b_enabled,
+            sprites_enabled: self.sprites_enabled,
+            window_enabled: self.window_enabled,
+            backdrop_enabled: self.backdrop_enabled,
         }
     }
 
@@ -293,6 +308,10 @@ impl GenesisEmulator {
     pub fn copy_vram(&self, out: &mut [Color], palette: u8, row_len: usize) {
         self.vdp.copy_vram(out, palette, row_len);
     }
+
+    pub fn dump_vdp_registers(&self, callback: impl FnMut(&str, &[(&str, &str)])) {
+        self.vdp.dump_registers(callback);
+    }
 }
 
 /// Render the current VDP frame buffer.
