@@ -14,7 +14,8 @@ use bincode::error::{DecodeError, EncodeError};
 use bincode::{BorrowDecode, Decode, Encode};
 use jgenesis_common::frontend::{Color, TimingMode};
 use jgenesis_common::num::{GetBit, U16Ext};
-use jgenesis_proc_macros::{EnumDisplay, EnumFromStr};
+use jgenesis_proc_macros::EnumDisplay;
+use std::fmt::{Display, Formatter};
 use z80_emu::traits::InterruptLine;
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq, Encode, Decode)]
@@ -78,7 +79,7 @@ impl ViewportSize {
     }
 }
 
-#[derive(Debug, Clone, Copy, PartialEq, Eq, Default, Encode, Decode, EnumDisplay, EnumFromStr)]
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Default, Encode, Decode, EnumDisplay)]
 #[cfg_attr(feature = "serde", derive(serde::Serialize, serde::Deserialize))]
 pub enum VdpVersion {
     NtscMasterSystem1,
@@ -169,6 +170,16 @@ enum Mode {
     GraphicsII,
 }
 
+impl Display for Mode {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        match self {
+            Self::Four => write!(f, "4"),
+            Self::Four224Line => write!(f, "4 (224-line)"),
+            Self::GraphicsII => write!(f, "Graphics II"),
+        }
+    }
+}
+
 impl Mode {
     fn from_mode_bits(mode_bits: [bool; 4]) -> Self {
         match mode_bits {
