@@ -3,6 +3,7 @@
 use clap::Parser;
 use env_logger::Env;
 use gb_core::api::{GbAspectRatio, GbPalette, GbcColorCorrection};
+use genesis_core::audio::LowPassFilter;
 use genesis_core::{GenesisAspectRatio, GenesisControllerType, GenesisRegion};
 use jgenesis_common::frontend::{EmulatorTrait, TimingMode};
 use jgenesis_native_config::AppConfig;
@@ -13,12 +14,13 @@ use jgenesis_native_driver::config::{
 };
 use jgenesis_native_driver::input::MappableInputs;
 use jgenesis_native_driver::{NativeEmulator, NativeTickEffect};
-use jgenesis_proc_macros::{EnumDisplay, EnumFromStr};
+use jgenesis_proc_macros::{CustomValueEnum, EnumAll, EnumDisplay};
 use jgenesis_renderer::config::{
     FilterMode, PreprocessShader, PrescaleFactor, Scanlines, VSyncMode, WgpuBackend,
 };
 use nes_core::api::NesAspectRatio;
 use s32x_core::api::S32XVideoOut;
+use segacd_core::api::PcmInterpolation;
 use smsgg_core::psg::Sn76489Version;
 use smsgg_core::{SmsModel, SmsRegion};
 use snes_core::api::{AudioInterpolationMode, SnesAspectRatio};
@@ -29,7 +31,7 @@ use std::hash::Hash;
 use std::num::{NonZeroU32, NonZeroU64};
 use std::path::{Path, PathBuf};
 
-#[derive(Debug, Clone, Copy, PartialEq, Eq, EnumDisplay, EnumFromStr)]
+#[derive(Debug, Clone, Copy, PartialEq, Eq, EnumAll, EnumDisplay, CustomValueEnum)]
 enum Hardware {
     MasterSystem,
     Genesis,
@@ -53,7 +55,7 @@ const HOTKEY_OPTIONS_HEADING: &str = "Hotkey Options";
 
 #[derive(Debug, Parser)]
 struct Args {
-    /// Hardware (MasterSystem / Genesis / SegaCd / Sega32X / Nes / Snes / GameBoy); defaults based on file extension if not set
+    /// Hardware; defaults based on file extension if not set
     #[arg(long)]
     hardware: Option<Hardware>,
 
@@ -69,7 +71,7 @@ struct Args {
     #[arg(long, value_name = "SLOT")]
     load_save_state: Option<usize>,
 
-    /// Force timing mode (Ntsc / Pal)
+    /// Force timing mode
     #[arg(long)]
     forced_timing_mode: Option<TimingMode>,
 
@@ -77,11 +79,11 @@ struct Args {
     #[arg(long)]
     remove_sprite_limit: Option<bool>,
 
-    /// Hide mouse cursor when over emulator window (Fullscreen / Always / Never)
+    /// Hide mouse cursor when over emulator window
     #[arg(long)]
     hide_mouse_cursor: Option<HideMouseCursor>,
 
-    /// Save file path (RomFolder / EmulatorFolder / Custom)
+    /// Save file path
     #[arg(long)]
     save_path: Option<ConfigSavePath>,
 
@@ -89,7 +91,7 @@ struct Args {
     #[arg(long)]
     custom_save_path: Option<PathBuf>,
 
-    /// Save state path (RomFolder / EmulatorFolder / Custom)
+    /// Save state path
     #[arg(long)]
     state_path: Option<ConfigSavePath>,
 
@@ -97,23 +99,23 @@ struct Args {
     #[arg(long)]
     custom_state_path: Option<PathBuf>,
 
-    /// MasterSystem model (Sms2 / Sms1)
+    /// MasterSystem model
     #[arg(long, help_heading = SMSGG_OPTIONS_HEADING)]
     sms_model: Option<SmsModel>,
 
-    /// Force PSG version (MasterSystem2 / Standard)
+    /// Force PSG version
     #[arg(long, help_heading = SMSGG_OPTIONS_HEADING)]
     psg_version: Option<Sn76489Version>,
 
-    /// Master System aspect ratio (Ntsc / Pal / SquarePixels / Stretched)
+    /// Master System aspect ratio
     #[arg(long, help_heading = SMSGG_OPTIONS_HEADING)]
     sms_aspect_ratio: Option<SmsAspectRatio>,
 
-    /// Game Gear aspect ratio (GgLcd / SquarePixels / Stretched)
+    /// Game Gear aspect ratio
     #[arg(long, help_heading = SMSGG_OPTIONS_HEADING)]
     gg_aspect_ratio: Option<GgAspectRatio>,
 
-    /// Master System region (International / Domestic)
+    /// Master System region
     #[arg(long, help_heading = SMSGG_OPTIONS_HEADING)]
     sms_region: Option<SmsRegion>,
 
@@ -163,6 +165,10 @@ struct Args {
     #[arg(long, help_heading = GENESIS_OPTIONS_HEADING)]
     emulate_ym2612_ladder_effect: Option<bool>,
 
+    /// Choose which low-pass filter to apply to Genesis audio output
+    #[arg(long, help_heading = GENESIS_OPTIONS_HEADING)]
+    genesis_low_pass_filter: Option<LowPassFilter>,
+
     /// Enable audio from the YM2612 FM chip
     #[arg(long, help_heading = GENESIS_OPTIONS_HEADING)]
     ym2612_enabled: Option<bool>,
@@ -171,7 +177,7 @@ struct Args {
     #[arg(long, help_heading = GENESIS_OPTIONS_HEADING)]
     genesis_psg_enabled: Option<bool>,
 
-    /// Aspect ratio (Ntsc / Pal / SquarePixels / Stretched)
+    /// Aspect ratio
     #[arg(long, help_heading = GENESIS_OPTIONS_HEADING)]
     genesis_aspect_ratio: Option<GenesisAspectRatio>,
 
@@ -183,11 +189,11 @@ struct Args {
     #[arg(long, help_heading = GENESIS_OPTIONS_HEADING)]
     genesis_region: Option<GenesisRegion>,
 
-    /// P1 Genesis controller type (ThreeButton / SixButton / None)
+    /// P1 Genesis controller type
     #[arg(long, help_heading = GENESIS_OPTIONS_HEADING)]
     genesis_p1_controller_type: Option<GenesisControllerType>,
 
-    /// P2 Genesis controller type (ThreeButton / SixButton / None)
+    /// P2 Genesis controller type
     #[arg(long, help_heading = GENESIS_OPTIONS_HEADING)]
     genesis_p2_controller_type: Option<GenesisControllerType>,
 
@@ -195,6 +201,10 @@ struct Args {
     #[arg(short = 'b', long, help_heading = SCD_OPTIONS_HEADING)]
     bios_path: Option<String>,
 
+    /// Sega CD PCM sound chip interpolation
+    #[arg(long, help_heading = SCD_OPTIONS_HEADING)]
+    scd_pcm_interpolation: Option<PcmInterpolation>,
+
     /// Enable Sega CD RAM cartridge mapping
     #[arg(long, help_heading = SCD_OPTIONS_HEADING)]
     enable_ram_cartridge: Option<bool>,
@@ -215,7 +225,7 @@ struct Args {
     #[arg(long, help_heading = SCD_OPTIONS_HEADING)]
     scd_cd_da_enabled: Option<bool>,
 
-    /// Set 32X video output (Combined / GenesisOnly / S32XOnly)
+    /// Set 32X video output
     #[arg(long, help_heading = S32X_OPTIONS_HEADING)]
     s32x_video_out: Option<S32XVideoOut>,
 
@@ -223,11 +233,11 @@ struct Args {
     #[arg(long, help_heading = S32X_OPTIONS_HEADING)]
     s32x_pwm_enabled: Option<bool>,
 
-    /// Aspect ratio (Ntsc / Pal / SquarePixels / Stretched)
+    /// Aspect ratio
     #[arg(long, help_heading = NES_OPTIONS_HEADING)]
     nes_aspect_ratio: Option<NesAspectRatio>,
 
-    /// NES P2 controller type (Gamepad / Zapper)
+    /// NES P2 controller type
     #[arg(long, help_heading = NES_OPTIONS_HEADING)]
     nes_p2_controller_type: Option<NesControllerType>,
 
@@ -263,11 +273,11 @@ struct Args {
     #[arg(long, help_heading = NES_OPTIONS_HEADING)]
     nes_audio_60hz_hack: Option<bool>,
 
-    /// SNES aspect ratio (Ntsc / Pal / SquarePixels / Stretched)
+    /// SNES aspect ratio
     #[arg(long, help_heading = SNES_OPTIONS_HEADING)]
     snes_aspect_ratio: Option<SnesAspectRatio>,
 
-    /// Audio interpolation mode (Gaussian / Hermite / Lagrange)
+    /// Audio interpolation mode
     #[arg(long, help_heading = SNES_OPTIONS_HEADING)]
     snes_audio_interpolation: Option<AudioInterpolationMode>,
 
@@ -279,7 +289,7 @@ struct Args {
     #[arg(long, help_heading = SNES_OPTIONS_HEADING)]
     gsu_overclock_factor: Option<NonZeroU64>,
 
-    /// Player 2 input device (Gamepad / SuperScope)
+    /// Player 2 input device
     #[arg(long, help_heading = SNES_OPTIONS_HEADING)]
     snes_p2_controller_type: Option<SnesControllerType>,
 
@@ -315,15 +325,15 @@ struct Args {
     #[arg(long, help_heading = GB_OPTIONS_HEADING)]
     pretend_to_be_gba: Option<bool>,
 
-    /// Aspect ratio (SquarePixels / Stretched)
+    /// Aspect ratio
     #[arg(long, help_heading = GB_OPTIONS_HEADING)]
     gb_aspect_ratio: Option<GbAspectRatio>,
 
-    /// Game Boy palette (BlackAndWhite / GreenTint / LimeGreen)
+    /// Game Boy palette
     #[arg(long, help_heading = GB_OPTIONS_HEADING)]
     gb_palette: Option<GbPalette>,
 
-    /// Game Boy Color color correction (None / GbcLcd / GbaLcd)
+    /// Game Boy Color color correction
     #[arg(long, help_heading = GB_OPTIONS_HEADING)]
     gbc_color_correction: Option<GbcColorCorrection>,
 
@@ -339,19 +349,23 @@ struct Args {
     #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
     window_height: Option<u32>,
 
+    /// Emulator window scale factor
+    #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
+    window_scale_factor: Option<f32>,
+
     /// Launch in fullscreen
     #[arg(long, default_value_t, help_heading = VIDEO_OPTIONS_HEADING)]
     fullscreen: bool,
 
-    /// Fullscreen mode (Borderless / Exclusive)
+    /// Fullscreen mode
     #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
     fullscreen_mode: Option<FullscreenMode>,
 
-    /// wgpu backend (Auto / Vulkan / DirectX12 / OpenGl)
+    /// wgpu backend
     #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
     wgpu_backend: Option<WgpuBackend>,
 
-    /// VSync mode (Enabled / Disabled / Fast)
+    /// VSync mode
     #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
     vsync_mode: Option<VSyncMode>,
 
@@ -367,7 +381,7 @@ struct Args {
     #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
     prescale_factor: Option<u32>,
 
-    /// Scanlines (None / Dim / Black)
+    /// Scanlines
     #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
     scanlines: Option<Scanlines>,
 
@@ -375,11 +389,11 @@ struct Args {
     #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
     force_integer_height_scaling: Option<bool>,
 
-    /// Filter mode (Nearest / Linear)
+    /// Filter mode
     #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
     filter_mode: Option<FilterMode>,
 
-    /// Preprocess shader (None / HorizontalBlurTwoPixels / HorizontalBlurThreePixels / HorizontalBlurSnesAdaptive / AntiDitherWeak / AntiDitherStrong)
+    /// Preprocess shader
     #[arg(long, help_heading = VIDEO_OPTIONS_HEADING)]
     preprocess_shader: Option<PreprocessShader>,
 
@@ -507,6 +521,7 @@ impl Args {
             genesis_render_vertical_border -> render_vertical_border,
             genesis_render_horizontal_border -> render_horizontal_border,
             quantize_ym2612_output,
+            genesis_low_pass_filter -> low_pass_filter,
             emulate_ym2612_ladder_effect,
             ym2612_enabled,
             genesis_psg_enabled -> psg_enabled,
@@ -529,6 +544,7 @@ impl Args {
 
         apply_overrides!(self, config.sega_cd, [
             enable_ram_cartridge,
+            scd_pcm_interpolation -> pcm_interpolation,
             scd_load_disc_into_ram -> load_disc_into_ram,
             scd_pcm_enabled -> pcm_enabled,
             scd_cd_da_enabled -> cd_audio_enabled,
@@ -598,6 +614,22 @@ impl Args {
         config.common.window_width = self.window_width;
         config.common.window_height = self.window_height;
 
+        match self.window_scale_factor {
+            Some(scale_factor) => {
+                config.common.window_scale_factor = Some(scale_factor);
+            }
+            None => {
+                if config.common.window_scale_factor.is_none() {
+                    let scale_factor = try_determine_scale_factor();
+                    config.common.window_scale_factor = scale_factor;
+
+                    if let Some(scale_factor) = scale_factor {
+                        log::info!("Detected scale factor of {scale_factor} using SDL2");
+                    }
+                }
+            }
+        }
+
         if self.fullscreen {
             config.common.launch_in_fullscreen = true;
         }
@@ -645,6 +677,12 @@ impl Args {
     }
 }
 
+fn try_determine_scale_factor() -> Option<f32> {
+    let sdl_ctx = sdl2::init().ok()?;
+    let video = sdl_ctx.video().ok()?;
+    jgenesis_native_driver::determine_scale_factor(&video, None)
+}
+
 fn main() -> anyhow::Result<()> {
     env_logger::Builder::from_env(
         Env::default().default_filter_or("info,wgpu_core=warn,wgpu_hal=warn"),
@@ -722,6 +760,9 @@ fn main() -> anyhow::Result<()> {
         log::error!("Unable to deserialize config file at '{}': {err}", config_path.display());
         AppConfig::default()
     });
+    if let Some(migrated_config) = jgenesis_native_config::migrate_config(&config, &config_str) {
+        config = migrated_config;
+    }
 
     args.apply_overrides(&mut config);
 
@@ -739,21 +780,13 @@ fn main() -> anyhow::Result<()> {
 fn run_sms(args: Args, config: AppConfig) -> anyhow::Result<()> {
     let mut emulator =
         jgenesis_native_driver::create_smsgg(config.smsgg_config(args.file_path.clone()))?;
-    init_emulator(&mut emulator, &args);
-
-    while emulator.render_frame()? != NativeTickEffect::Exit {}
-
-    Ok(())
+    run_emulator(&mut emulator, &args)
 }
 
 fn run_genesis(args: Args, config: AppConfig) -> anyhow::Result<()> {
     let mut emulator =
         jgenesis_native_driver::create_genesis(config.genesis_config(args.file_path.clone()))?;
-    init_emulator(&mut emulator, &args);
-
-    while emulator.render_frame()? != NativeTickEffect::Exit {}
-
-    Ok(())
+    run_emulator(&mut emulator, &args)
 }
 
 fn run_sega_cd(args: Args, config: AppConfig) -> anyhow::Result<()> {
@@ -761,66 +794,54 @@ fn run_sega_cd(args: Args, config: AppConfig) -> anyhow::Result<()> {
     scd_config.run_without_disc = args.scd_no_disc;
 
     let mut emulator = jgenesis_native_driver::create_sega_cd(scd_config)?;
-    init_emulator(&mut emulator, &args);
-
-    while emulator.render_frame()? != NativeTickEffect::Exit {}
-
-    Ok(())
+    run_emulator(&mut emulator, &args)
 }
 
 fn run_32x(args: Args, config: AppConfig) -> anyhow::Result<()> {
     let mut emulator =
         jgenesis_native_driver::create_32x(config.sega_32x_config(args.file_path.clone()))?;
-    init_emulator(&mut emulator, &args);
-
-    while emulator.render_frame()? != NativeTickEffect::Exit {}
-
-    Ok(())
+    run_emulator(&mut emulator, &args)
 }
 
 fn run_nes(args: Args, config: AppConfig) -> anyhow::Result<()> {
     let mut emulator =
         jgenesis_native_driver::create_nes(config.nes_config(args.file_path.clone()))?;
-    init_emulator(&mut emulator, &args);
-
-    while emulator.render_frame()? != NativeTickEffect::Exit {}
-
-    Ok(())
+    run_emulator(&mut emulator, &args)
 }
 
 fn run_snes(args: Args, config: AppConfig) -> anyhow::Result<()> {
     let mut emulator =
         jgenesis_native_driver::create_snes(config.snes_config(args.file_path.clone()))?;
-    init_emulator(&mut emulator, &args);
-
-    while emulator.render_frame()? != NativeTickEffect::Exit {}
-
-    Ok(())
+    run_emulator(&mut emulator, &args)
 }
 
 fn run_gb(args: Args, config: AppConfig) -> anyhow::Result<()> {
     let mut emulator = jgenesis_native_driver::create_gb(config.gb_config(args.file_path.clone()))?;
-    init_emulator(&mut emulator, &args);
-
-    while emulator.render_frame()? != NativeTickEffect::Exit {}
-
-    Ok(())
+    run_emulator(&mut emulator, &args)
 }
 
 // TODO simplify these trait bounds (required by load_state() definition)
-fn init_emulator<Inputs, Button, Config, Emulator>(
+fn run_emulator<Inputs, Button, Config, Emulator>(
     emulator: &mut NativeEmulator<Inputs, Button, Config, Emulator>,
     args: &Args,
-) where
+) -> anyhow::Result<()>
+where
     Inputs: Default + MappableInputs<Button>,
     Button: Debug + Copy + Hash + Eq,
     Emulator: EmulatorTrait<Inputs = Inputs, Config = Config>,
 {
-    let Some(save_state_slot) = args.load_save_state else { return };
+    if let Some(save_state_slot) = args.load_save_state {
+        log::info!("Loading save state slot {save_state_slot} at launch");
 
-    log::info!("Loading save state slot {save_state_slot} at launch");
+        if let Err(err) = emulator.load_state(save_state_slot) {
+            log::error!("Error loading save state slot {save_state_slot} at launch: {err}");
+        }
+    }
 
-    if let Err(err) = emulator.load_state(save_state_slot) {
-        log::error!("Error loading save state slot {save_state_slot} at launch: {err}");
+    loop {
+        match emulator.render_frame()? {
+            Some(NativeTickEffect::PowerOff | NativeTickEffect::Exit) => return Ok(()),
+            None => {}
+        }
     }
 }
