@@ -10,7 +10,7 @@ mod snes;
 use crate::app::common::SavePathSelect;
 use crate::app::input::{GenericButton, InputMappingSet};
 use crate::app::nes::OverscanState;
-use crate::app::romlist::{Console, RomListThreadHandle, RomMetadata};
+use crate::app::romlist::{RomListThreadHandle, RomMetadata};
 use crate::emuthread;
 use crate::emuthread::{EmuThreadCommand, EmuThreadHandle, EmuThreadStatus};
 use eframe::Frame;
@@ -23,6 +23,7 @@ use egui::{
 use egui_extras::{Column, TableBuilder};
 use jgenesis_native_config::{AppConfig, EguiTheme, ListFilters, RecentOpen};
 use jgenesis_native_driver::config::HideMouseCursor;
+use jgenesis_proc_macros::{EnumAll, EnumDisplay, EnumFromStr};
 use jgenesis_renderer::config::Scanlines;
 use rfd::FileDialog;
 use std::collections::{HashMap, HashSet};
@@ -35,6 +36,19 @@ use time::util::local_offset;
 use time::util::local_offset::Soundness;
 use time::{OffsetDateTime, UtcOffset, format_description};
 
+#[derive(Debug, Clone, Copy, PartialEq, Eq, EnumAll, EnumDisplay, EnumFromStr)]
+pub enum Console {
+    MasterSystem,
+    GameGear,
+    Genesis,
+    SegaCd,
+    Sega32X,
+    Nes,
+    Snes,
+    GameBoy,
+    GameBoyColor,
+}
+
 trait ListFiltersExt {
     fn to_console_vec(&self) -> Vec<Console>;
 
@@ -255,14 +269,34 @@ pub struct App {
     load_at_startup: Option<LoadAtStartup>,
 }
 
+fn load_app_config(config_path: &Path) -> AppConfig {
+    let config_str = fs::read_to_string(config_path).unwrap_or_default();
+    let mut config = toml::from_str(&config_str).unwrap_or_else(|err| {
+        log::error!("Error deserializing app config at '{}': {err}", config_path.display());
+        AppConfig::default()
+    });
+
+    if let Some(migrated_config) = jgenesis_native_config::migrate_config(&config, &config_str) {
+        if config != migrated_config {
+            config = migrated_config;
+            if let Err(err) = fs::write(config_path, toml::to_string_pretty(&config).unwrap()) {
+                log::error!("Error serializing app config: {err}");
+            }
+        }
+    }
+
+    config
+}
+
 impl App {
     #[must_use]
     pub fn new(config_path: PathBuf, load_at_startup: Option<LoadAtStartup>, ctx: Context) -> Self {
-        let config = AppConfig::from_file(&config_path);
-        let state = AppState::from_config(&config);
-        let emu_thread = emuthread::spawn(ctx);
+        let config = load_app_config(&config_path);
 
-        let rom_list_thread = RomListThreadHandle::spawn(Arc::clone(&state.rom_list));
+        let state = AppState::from_config(&config);
+        let emu_thread = emuthread::spawn(ctx.clone());
+
+        let rom_list_thread = RomListThreadHandle::spawn(Arc::clone(&state.rom_list), ctx);
         rom_list_thread.request_scan(config.rom_search_dirs.clone());
 
         Self { config, state, config_path, emu_thread, rom_list_thread, load_at_startup }
@@ -322,50 +356,12 @@ impl App {
         self.config.recent_open_list.truncate(10);
         self.state.recent_open_list = romlist::from_recent_opens(&self.config.recent_open_list);
 
-        match console {
-            Console::MasterSystem | Console::GameGear => {
-                self.emu_thread.stop_emulator_if_running();
-
-                let config = self.config.smsgg_config(path);
-                self.emu_thread.send(EmuThreadCommand::RunSms(config));
-            }
-            Console::Genesis => {
-                self.emu_thread.stop_emulator_if_running();
-
-                let config = self.config.genesis_config(path);
-                self.emu_thread.send(EmuThreadCommand::RunGenesis(config));
-            }
-            Console::SegaCd => {
-                self.emu_thread.stop_emulator_if_running();
-
-                let config = self.config.sega_cd_config(path);
-                self.emu_thread.send(EmuThreadCommand::RunSegaCd(config));
-            }
-            Console::Sega32X => {
-                self.emu_thread.stop_emulator_if_running();
-
-                let config = self.config.sega_32x_config(path);
-                self.emu_thread.send(EmuThreadCommand::Run32X(config));
-            }
-            Console::Nes => {
-                self.emu_thread.stop_emulator_if_running();
-
-                let config = self.config.nes_config(path);
-                self.emu_thread.send(EmuThreadCommand::RunNes(config));
-            }
-            Console::Snes => {
-                self.emu_thread.stop_emulator_if_running();
-
-                let config = self.config.snes_config(path);
-                self.emu_thread.send(EmuThreadCommand::RunSnes(config));
-            }
-            Console::GameBoy | Console::GameBoyColor => {
-                self.emu_thread.stop_emulator_if_running();
-
-                let config = self.config.gb_config(path);
-                self.emu_thread.send(EmuThreadCommand::RunGameBoy(config));
-            }
-        }
+        self.emu_thread.stop_emulator_if_running();
+        self.emu_thread.send(EmuThreadCommand::Run {
+            console,
+            config: Box::new(self.config.clone()),
+            file_path: path,
+        });
     }
 
     fn add_rom_search_directory(&mut self) {
@@ -1057,16 +1053,10 @@ impl App {
     }
 
     fn reload_config(&mut self) {
-        // TODO this is terrible; should only generate and send the config for the currently-running emulator
-        self.emu_thread.reload_config(
-            self.config.smsgg_config(self.state.current_file_path.clone()),
-            self.config.genesis_config(self.state.current_file_path.clone()),
-            self.config.sega_cd_config(self.state.current_file_path.clone()),
-            self.config.sega_32x_config(self.state.current_file_path.clone()),
-            self.config.nes_config(self.state.current_file_path.clone()),
-            self.config.snes_config(self.state.current_file_path.clone()),
-            self.config.gb_config(self.state.current_file_path.clone()),
-        );
+        self.emu_thread.send(EmuThreadCommand::ReloadConfig(
+            Box::new(self.config.clone()),
+            self.state.current_file_path.clone(),
+        ));
     }
 
     fn refresh_filtered_rom_list(&mut self) {
@@ -1084,6 +1074,11 @@ impl App {
 
 impl eframe::App for App {
     fn update(&mut self, ctx: &Context, _frame: &mut Frame) {
+        if self.emu_thread.exit_signal() {
+            ctx.send_viewport_cmd(ViewportCommand::Close);
+            return;
+        }
+
         if self.state.rom_list_refresh_needed && !self.rom_list_thread.any_scans_in_progress() {
             self.state.rom_list_refresh_needed = false;
             self.refresh_filtered_rom_list();
