@@ -1,10 +1,9 @@
 mod inputwindow;
 
+use crate::app::Console;
 use crate::emuthread::inputwindow::InputWindow;
 use anyhow::anyhow;
-use jgenesis_native_driver::config::{
-    GameBoyConfig, GenesisConfig, NesConfig, Sega32XConfig, SegaCdConfig, SmsGgConfig, SnesConfig,
-};
+use jgenesis_native_config::AppConfig;
 use jgenesis_native_driver::input::{
     AxisDirection, GamepadAction, GenericInput, HatDirection, Joysticks,
 };
@@ -15,10 +14,11 @@ use jgenesis_native_driver::{
 };
 use sdl2::EventPump;
 use sdl2::event::Event;
-use sdl2::joystick::HatState;
+use sdl2::joystick::{HatState, Joystick};
 use segacd_core::api::SegaCdLoadResult;
+use std::collections::HashSet;
 use std::path::PathBuf;
-use std::sync::atomic::{AtomicU8, Ordering};
+use std::sync::atomic::{AtomicBool, AtomicU8, Ordering};
 use std::sync::mpsc::{Receiver, Sender, TryRecvError};
 use std::sync::{Arc, Mutex, MutexGuard, mpsc};
 use std::thread;
@@ -67,22 +67,24 @@ impl EmuThreadStatus {
     }
 }
 
+impl Console {
+    fn running_status(self) -> EmuThreadStatus {
+        match self {
+            Self::MasterSystem | Self::GameGear => EmuThreadStatus::RunningSmsGg,
+            Self::Genesis => EmuThreadStatus::RunningGenesis,
+            Self::SegaCd => EmuThreadStatus::RunningSegaCd,
+            Self::Sega32X => EmuThreadStatus::Running32X,
+            Self::Nes => EmuThreadStatus::RunningNes,
+            Self::Snes => EmuThreadStatus::RunningSnes,
+            Self::GameBoy | Self::GameBoyColor => EmuThreadStatus::RunningGameBoy,
+        }
+    }
+}
+
 #[derive(Debug, Clone)]
 pub enum EmuThreadCommand {
-    RunSms(Box<SmsGgConfig>),
-    RunGenesis(Box<GenesisConfig>),
-    RunSegaCd(Box<SegaCdConfig>),
-    Run32X(Box<Sega32XConfig>),
-    RunNes(Box<NesConfig>),
-    RunSnes(Box<SnesConfig>),
-    RunGameBoy(Box<GameBoyConfig>),
-    ReloadSmsGgConfig(Box<SmsGgConfig>),
-    ReloadGenesisConfig(Box<GenesisConfig>),
-    ReloadSegaCdConfig(Box<SegaCdConfig>),
-    Reload32XConfig(Box<Sega32XConfig>),
-    ReloadNesConfig(Box<NesConfig>),
-    ReloadSnesConfig(Box<SnesConfig>),
-    ReloadGameBoyConfig(Box<GameBoyConfig>),
+    Run { console: Console, config: Box<AppConfig>, file_path: String },
+    ReloadConfig(Box<AppConfig>, String),
     StopEmulator,
     CollectInput { axis_deadzone: i16 },
     SoftReset,
@@ -100,6 +102,7 @@ pub struct EmuThreadHandle {
     input_receiver: Receiver<Option<Vec<GenericInput>>>,
     save_state_metadata: Arc<Mutex<SaveStateMetadata>>,
     emulator_error: Arc<Mutex<Option<anyhow::Error>>>,
+    exit_signal: Arc<AtomicBool>,
 }
 
 impl EmuThreadHandle {
@@ -138,235 +141,92 @@ impl EmuThreadHandle {
         }
     }
 
-    // TODO fix this
-    #[allow(clippy::too_many_arguments)]
-    pub fn reload_config(
-        &self,
-        smsgg_config: Box<SmsGgConfig>,
-        genesis_config: Box<GenesisConfig>,
-        sega_cd_config: Box<SegaCdConfig>,
-        s32x_config: Box<Sega32XConfig>,
-        nes_config: Box<NesConfig>,
-        snes_config: Box<SnesConfig>,
-        gb_config: Box<GameBoyConfig>,
-    ) {
-        match self.status() {
-            EmuThreadStatus::RunningSmsGg => {
-                self.send(EmuThreadCommand::ReloadSmsGgConfig(smsgg_config));
-            }
-            EmuThreadStatus::RunningGenesis => {
-                self.send(EmuThreadCommand::ReloadGenesisConfig(genesis_config));
-            }
-            EmuThreadStatus::RunningSegaCd => {
-                self.send(EmuThreadCommand::ReloadSegaCdConfig(sega_cd_config));
-            }
-            EmuThreadStatus::Running32X => {
-                self.send(EmuThreadCommand::Reload32XConfig(s32x_config));
-            }
-            EmuThreadStatus::RunningNes => {
-                self.send(EmuThreadCommand::ReloadNesConfig(nes_config));
-            }
-            EmuThreadStatus::RunningSnes => {
-                self.send(EmuThreadCommand::ReloadSnesConfig(snes_config));
-            }
-            EmuThreadStatus::RunningGameBoy => {
-                self.send(EmuThreadCommand::ReloadGameBoyConfig(gb_config));
-            }
-            EmuThreadStatus::Idle | EmuThreadStatus::WaitingForFirstCommand => {}
-        }
+    pub fn exit_signal(&self) -> bool {
+        self.exit_signal.load(Ordering::Relaxed)
     }
 }
 
-pub fn spawn(ctx: egui::Context) -> EmuThreadHandle {
+pub fn spawn(egui_ctx: egui::Context) -> EmuThreadHandle {
     let status = Arc::new(AtomicU8::new(EmuThreadStatus::WaitingForFirstCommand as u8));
     let (command_sender, command_receiver) = mpsc::channel();
     let (input_sender, input_receiver) = mpsc::channel();
     let save_state_metadata = Arc::new(Mutex::new(SaveStateMetadata::default()));
     let emulator_error = Arc::new(Mutex::new(None));
+    let exit_signal = Arc::new(AtomicBool::new(false));
 
     {
         let status = Arc::clone(&status);
         let save_state_metadata = Arc::clone(&save_state_metadata);
         let emulator_error = Arc::clone(&emulator_error);
+        let exit_signal = Arc::clone(&exit_signal);
         thread::spawn(move || {
-            thread_run(
-                ctx,
+            thread_run(EmuThreadContext {
+                egui_ctx,
                 command_receiver,
                 input_sender,
                 status,
                 save_state_metadata,
                 emulator_error,
-            );
+                exit_signal,
+            });
         });
     }
 
-    EmuThreadHandle { status, command_sender, input_receiver, save_state_metadata, emulator_error }
+    EmuThreadHandle {
+        status,
+        command_sender,
+        input_receiver,
+        save_state_metadata,
+        emulator_error,
+        exit_signal,
+    }
 }
 
-fn thread_run(
-    ctx: egui::Context,
+struct EmuThreadContext {
+    egui_ctx: egui::Context,
     command_receiver: Receiver<EmuThreadCommand>,
     input_sender: Sender<Option<Vec<GenericInput>>>,
     status: Arc<AtomicU8>,
     save_state_metadata: Arc<Mutex<SaveStateMetadata>>,
     emulator_error: Arc<Mutex<Option<anyhow::Error>>>,
-) {
+    exit_signal: Arc<AtomicBool>,
+}
+
+fn thread_run(ctx: EmuThreadContext) {
     loop {
-        if status.load(Ordering::Relaxed) != EmuThreadStatus::WaitingForFirstCommand as u8 {
-            status.store(EmuThreadStatus::Idle as u8, Ordering::Relaxed);
+        if ctx.status.load(Ordering::Relaxed) != EmuThreadStatus::WaitingForFirstCommand as u8 {
+            ctx.status.store(EmuThreadStatus::Idle as u8, Ordering::Relaxed);
         }
 
         // Force a repaint at the start of the loop so that the GUI will always repaint after an
         // emulator exits. This will immediately display the error window if there was an
         // error, and it will also force quit immediately if auto-close is enabled
-        ctx.request_repaint();
+        ctx.egui_ctx.request_repaint();
 
-        match command_receiver.recv() {
-            Ok(EmuThreadCommand::RunSms(config)) => {
-                status.store(EmuThreadStatus::RunningSmsGg as u8, Ordering::Relaxed);
+        match ctx.command_receiver.recv() {
+            Ok(EmuThreadCommand::Run { console, mut config, file_path }) => {
+                ctx.status.store(console.running_status() as u8, Ordering::Relaxed);
 
-                let emulator = match jgenesis_native_driver::create_smsgg(config) {
+                if let Some(native_ppi) = ctx.egui_ctx.native_pixels_per_point() {
+                    log::info!("Setting emulator window scale factor to {native_ppi}");
+                    config.common.window_scale_factor = Some(native_ppi);
+                }
+
+                let emulator = match GenericEmulator::create(console, config, file_path) {
                     Ok(emulator) => emulator,
                     Err(err) => {
-                        log::error!("Error initializing SMS/GG emulator: {err}");
-                        *emulator_error.lock().unwrap() = Some(err.into());
+                        log::error!("Error initializing emulator: {err}");
+                        *ctx.emulator_error.lock().unwrap() = Some(err.into());
                         continue;
                     }
                 };
-                run_emulator(
-                    GenericEmulator::SmsGg(emulator),
-                    &command_receiver,
-                    &input_sender,
-                    &save_state_metadata,
-                    &emulator_error,
-                    &ctx,
-                );
-            }
-            Ok(EmuThreadCommand::RunGenesis(config)) => {
-                status.store(EmuThreadStatus::RunningGenesis as u8, Ordering::Relaxed);
-
-                let emulator = match jgenesis_native_driver::create_genesis(config) {
-                    Ok(emulator) => emulator,
-                    Err(err) => {
-                        log::error!("Error initializing Genesis emulator: {err}");
-                        *emulator_error.lock().unwrap() = Some(err.into());
-                        continue;
-                    }
-                };
-                run_emulator(
-                    GenericEmulator::Genesis(emulator),
-                    &command_receiver,
-                    &input_sender,
-                    &save_state_metadata,
-                    &emulator_error,
-                    &ctx,
-                );
-            }
-            Ok(EmuThreadCommand::RunSegaCd(config)) => {
-                status.store(EmuThreadStatus::RunningSegaCd as u8, Ordering::Relaxed);
-
-                let emulator = match jgenesis_native_driver::create_sega_cd(config) {
-                    Ok(emulator) => emulator,
-                    Err(err) => {
-                        log::error!("Error initializing Sega CD emulator: {err}");
-                        *emulator_error.lock().unwrap() = Some(err.into());
-                        continue;
-                    }
-                };
-                run_emulator(
-                    GenericEmulator::SegaCd(emulator),
-                    &command_receiver,
-                    &input_sender,
-                    &save_state_metadata,
-                    &emulator_error,
-                    &ctx,
-                );
-            }
-            Ok(EmuThreadCommand::Run32X(config)) => {
-                status.store(EmuThreadStatus::Running32X as u8, Ordering::Relaxed);
-
-                let emulator = match jgenesis_native_driver::create_32x(config) {
-                    Ok(emulator) => emulator,
-                    Err(err) => {
-                        log::error!("Error initializing 32X emulator: {err}");
-                        *emulator_error.lock().unwrap() = Some(err.into());
-                        continue;
-                    }
-                };
-                run_emulator(
-                    GenericEmulator::Sega32X(emulator),
-                    &command_receiver,
-                    &input_sender,
-                    &save_state_metadata,
-                    &emulator_error,
-                    &ctx,
-                );
-            }
-            Ok(EmuThreadCommand::RunNes(config)) => {
-                status.store(EmuThreadStatus::RunningNes as u8, Ordering::Relaxed);
-
-                let emulator = match jgenesis_native_driver::create_nes(config) {
-                    Ok(emulator) => emulator,
-                    Err(err) => {
-                        log::error!("Error initializing NES emulator: {err}");
-                        *emulator_error.lock().unwrap() = Some(err.into());
-                        continue;
-                    }
-                };
-                run_emulator(
-                    GenericEmulator::Nes(emulator),
-                    &command_receiver,
-                    &input_sender,
-                    &save_state_metadata,
-                    &emulator_error,
-                    &ctx,
-                );
-            }
-            Ok(EmuThreadCommand::RunSnes(config)) => {
-                status.store(EmuThreadStatus::RunningSnes as u8, Ordering::Relaxed);
-
-                let emulator = match jgenesis_native_driver::create_snes(config) {
-                    Ok(emulator) => emulator,
-                    Err(err) => {
-                        log::error!("Error initializing SNES emulator: {err}");
-                        *emulator_error.lock().unwrap() = Some(err.into());
-                        continue;
-                    }
-                };
-                run_emulator(
-                    GenericEmulator::Snes(emulator),
-                    &command_receiver,
-                    &input_sender,
-                    &save_state_metadata,
-                    &emulator_error,
-                    &ctx,
-                );
-            }
-            Ok(EmuThreadCommand::RunGameBoy(config)) => {
-                status.store(EmuThreadStatus::RunningGameBoy as u8, Ordering::Relaxed);
-
-                let emulator = match jgenesis_native_driver::create_gb(config) {
-                    Ok(emulator) => emulator,
-                    Err(err) => {
-                        log::error!("Error initializing Game Boy emulator: {err}");
-                        *emulator_error.lock().unwrap() = Some(err.into());
-                        continue;
-                    }
-                };
-                run_emulator(
-                    GenericEmulator::GameBoy(emulator),
-                    &command_receiver,
-                    &input_sender,
-                    &save_state_metadata,
-                    &emulator_error,
-                    &ctx,
-                );
+                run_emulator(emulator, &ctx);
             }
             Ok(EmuThreadCommand::CollectInput { axis_deadzone }) => {
-                match collect_input_not_running(axis_deadzone, ctx.pixels_per_point()) {
+                match collect_input_not_running(axis_deadzone, ctx.egui_ctx.pixels_per_point()) {
                     Ok(input) => {
-                        input_sender.send(input).unwrap();
-                        ctx.request_repaint();
+                        ctx.input_sender.send(input).unwrap();
+                        ctx.egui_ctx.request_repaint();
                     }
                     Err(err) => {
                         log::error!("Error collecting SDL2 input: {err}");
@@ -375,13 +235,7 @@ fn thread_run(
             }
             Ok(
                 EmuThreadCommand::StopEmulator
-                | EmuThreadCommand::ReloadSmsGgConfig(_)
-                | EmuThreadCommand::ReloadGenesisConfig(_)
-                | EmuThreadCommand::ReloadSegaCdConfig(_)
-                | EmuThreadCommand::Reload32XConfig(_)
-                | EmuThreadCommand::ReloadNesConfig(_)
-                | EmuThreadCommand::ReloadSnesConfig(_)
-                | EmuThreadCommand::ReloadGameBoyConfig(_)
+                | EmuThreadCommand::ReloadConfig(..)
                 | EmuThreadCommand::SoftReset
                 | EmuThreadCommand::HardReset
                 | EmuThreadCommand::OpenMemoryViewer
@@ -425,60 +279,46 @@ macro_rules! match_each_emulator_variant {
 }
 
 impl GenericEmulator {
-    fn reload_smsgg_config(&mut self, config: Box<SmsGgConfig>) -> Result<(), AudioError> {
-        if let Self::SmsGg(emulator) = self {
-            emulator.reload_smsgg_config(config)?;
-        }
+    fn create(
+        console: Console,
+        config: Box<AppConfig>,
+        path: String,
+    ) -> NativeEmulatorResult<Self> {
+        let emulator = match console {
+            Console::MasterSystem | Console::GameGear => {
+                Self::SmsGg(jgenesis_native_driver::create_smsgg(config.smsgg_config(path))?)
+            }
+            Console::Genesis => {
+                Self::Genesis(jgenesis_native_driver::create_genesis(config.genesis_config(path))?)
+            }
+            Console::SegaCd => {
+                Self::SegaCd(jgenesis_native_driver::create_sega_cd(config.sega_cd_config(path))?)
+            }
+            Console::Sega32X => {
+                Self::Sega32X(jgenesis_native_driver::create_32x(config.sega_32x_config(path))?)
+            }
+            Console::Nes => Self::Nes(jgenesis_native_driver::create_nes(config.nes_config(path))?),
+            Console::Snes => {
+                Self::Snes(jgenesis_native_driver::create_snes(config.snes_config(path))?)
+            }
+            Console::GameBoy | Console::GameBoyColor => {
+                Self::GameBoy(jgenesis_native_driver::create_gb(config.gb_config(path))?)
+            }
+        };
 
-        Ok(())
+        Ok(emulator)
     }
 
-    fn reload_genesis_config(&mut self, config: Box<GenesisConfig>) -> Result<(), AudioError> {
-        if let Self::Genesis(emulator) = self {
-            emulator.reload_genesis_config(config)?;
+    fn reload_config(&mut self, config: Box<AppConfig>, path: String) -> Result<(), AudioError> {
+        match self {
+            Self::SmsGg(emulator) => emulator.reload_smsgg_config(config.smsgg_config(path)),
+            Self::Genesis(emulator) => emulator.reload_genesis_config(config.genesis_config(path)),
+            Self::SegaCd(emulator) => emulator.reload_sega_cd_config(config.sega_cd_config(path)),
+            Self::Sega32X(emulator) => emulator.reload_32x_config(config.sega_32x_config(path)),
+            Self::Nes(emulator) => emulator.reload_nes_config(config.nes_config(path)),
+            Self::Snes(emulator) => emulator.reload_snes_config(config.snes_config(path)),
+            Self::GameBoy(emulator) => emulator.reload_gb_config(config.gb_config(path)),
         }
-
-        Ok(())
-    }
-
-    fn reload_sega_cd_config(&mut self, config: Box<SegaCdConfig>) -> Result<(), AudioError> {
-        if let Self::SegaCd(emulator) = self {
-            emulator.reload_sega_cd_config(config)?;
-        }
-
-        Ok(())
-    }
-
-    fn reload_32x_config(&mut self, config: Box<Sega32XConfig>) -> Result<(), AudioError> {
-        if let Self::Sega32X(emulator) = self {
-            emulator.reload_32x_config(config)?;
-        }
-
-        Ok(())
-    }
-
-    fn reload_nes_config(&mut self, config: Box<NesConfig>) -> Result<(), AudioError> {
-        if let Self::Nes(emulator) = self {
-            emulator.reload_nes_config(config)?;
-        }
-
-        Ok(())
-    }
-
-    fn reload_snes_config(&mut self, config: Box<SnesConfig>) -> Result<(), AudioError> {
-        if let Self::Snes(emulator) = self {
-            emulator.reload_snes_config(config)?;
-        }
-
-        Ok(())
-    }
-
-    fn reload_gb_config(&mut self, config: Box<GameBoyConfig>) -> Result<(), AudioError> {
-        if let Self::GameBoy(emulator) = self {
-            emulator.reload_gb_config(config)?;
-        }
-
-        Ok(())
     }
 
     fn remove_disc(&mut self) {
@@ -495,7 +335,7 @@ impl GenericEmulator {
         Ok(())
     }
 
-    fn render_frame(&mut self) -> NativeEmulatorResult<NativeTickEffect> {
+    fn render_frame(&mut self) -> NativeEmulatorResult<Option<NativeTickEffect>> {
         match_each_emulator_variant!(self, emulator => emulator.render_frame())
     }
 
@@ -538,60 +378,17 @@ impl GenericEmulator {
     }
 }
 
-fn run_emulator(
-    mut emulator: GenericEmulator,
-    command_receiver: &Receiver<EmuThreadCommand>,
-    input_sender: &Sender<Option<Vec<GenericInput>>>,
-    save_state_metadata: &Arc<Mutex<SaveStateMetadata>>,
-    emulator_error: &Arc<Mutex<Option<anyhow::Error>>>,
-    ctx: &egui::Context,
-) {
+fn run_emulator(mut emulator: GenericEmulator, ctx: &EmuThreadContext) {
     loop {
         match emulator.render_frame() {
-            Ok(NativeTickEffect::None) => {
-                *save_state_metadata.lock().unwrap() = emulator.save_state_metadata();
+            Ok(None) => {
+                *ctx.save_state_metadata.lock().unwrap() = emulator.save_state_metadata();
 
-                while let Ok(command) = command_receiver.try_recv() {
+                while let Ok(command) = ctx.command_receiver.try_recv() {
                     match command {
-                        EmuThreadCommand::ReloadSmsGgConfig(config) => {
-                            if let Err(err) = emulator.reload_smsgg_config(config) {
-                                *emulator_error.lock().unwrap() = Some(err.into());
-                                return;
-                            }
-                        }
-                        EmuThreadCommand::ReloadGenesisConfig(config) => {
-                            if let Err(err) = emulator.reload_genesis_config(config) {
-                                *emulator_error.lock().unwrap() = Some(err.into());
-                                return;
-                            }
-                        }
-                        EmuThreadCommand::ReloadSegaCdConfig(config) => {
-                            if let Err(err) = emulator.reload_sega_cd_config(config) {
-                                *emulator_error.lock().unwrap() = Some(err.into());
-                                return;
-                            }
-                        }
-                        EmuThreadCommand::Reload32XConfig(config) => {
-                            if let Err(err) = emulator.reload_32x_config(config) {
-                                *emulator_error.lock().unwrap() = Some(err.into());
-                                return;
-                            }
-                        }
-                        EmuThreadCommand::ReloadNesConfig(config) => {
-                            if let Err(err) = emulator.reload_nes_config(config) {
-                                *emulator_error.lock().unwrap() = Some(err.into());
-                                return;
-                            }
-                        }
-                        EmuThreadCommand::ReloadSnesConfig(config) => {
-                            if let Err(err) = emulator.reload_snes_config(config) {
-                                *emulator_error.lock().unwrap() = Some(err.into());
-                                return;
-                            }
-                        }
-                        EmuThreadCommand::ReloadGameBoyConfig(config) => {
-                            if let Err(err) = emulator.reload_gb_config(config) {
-                                *emulator_error.lock().unwrap() = Some(err.into());
+                        EmuThreadCommand::ReloadConfig(config, path) => {
+                            if let Err(err) = emulator.reload_config(config, path) {
+                                *ctx.emulator_error.lock().unwrap() = Some(err.into());
                                 return;
                             }
                         }
@@ -609,8 +406,8 @@ fn run_emulator(
                             let is_none = input.is_none();
 
                             log::debug!("Sending collect input result {input:?}");
-                            input_sender.send(input).unwrap();
-                            ctx.request_repaint();
+                            ctx.input_sender.send(input).unwrap();
+                            ctx.egui_ctx.request_repaint();
 
                             if is_none {
                                 // Window was closed
@@ -625,26 +422,24 @@ fn run_emulator(
                         EmuThreadCommand::SegaCdRemoveDisc => emulator.remove_disc(),
                         EmuThreadCommand::SegaCdChangeDisc(path) => {
                             if let Err(err) = emulator.change_disc(path) {
-                                *emulator_error.lock().unwrap() = Some(err.into());
+                                *ctx.emulator_error.lock().unwrap() = Some(err.into());
                                 return;
                             }
                         }
-                        EmuThreadCommand::RunSms(_)
-                        | EmuThreadCommand::RunGenesis(_)
-                        | EmuThreadCommand::RunSegaCd(_)
-                        | EmuThreadCommand::Run32X(_)
-                        | EmuThreadCommand::RunNes(_)
-                        | EmuThreadCommand::RunSnes(_)
-                        | EmuThreadCommand::RunGameBoy(_) => {}
+                        EmuThreadCommand::Run { .. } => {}
                     }
                 }
             }
-            Ok(NativeTickEffect::Exit) => {
+            Ok(Some(NativeTickEffect::PowerOff)) => {
+                return;
+            }
+            Ok(Some(NativeTickEffect::Exit)) => {
+                ctx.exit_signal.store(true, Ordering::Relaxed);
                 return;
             }
             Err(err) => {
                 log::error!("Emulator terminated with an error: {err}");
-                *emulator_error.lock().unwrap() = Some(err.into());
+                *ctx.emulator_error.lock().unwrap() = Some(err.into());
                 return;
             }
         }
@@ -701,13 +496,85 @@ impl VecSet {
     }
 }
 
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+enum CollectionDone {
+    No,
+    Yes,
+}
+
+struct CollectedInputs {
+    inputs: VecSet,
+    gamepad_starting_states: HashSet<GenericInput>,
+}
+
+impl CollectedInputs {
+    fn new(joysticks: &Joysticks, axis_deadzone: i16) -> Self {
+        let gamepad_starting_states = joysticks
+            .all_devices()
+            .flat_map(|(device_id, joystick)| {
+                joystick_starting_state(device_id, joystick, axis_deadzone)
+            })
+            .collect();
+
+        Self { inputs: VecSet::new(), gamepad_starting_states }
+    }
+
+    fn add_device(&mut self, device_id: u32, joystick: &Joystick, axis_deadzone: i16) {
+        self.gamepad_starting_states.extend(joystick_starting_state(
+            device_id,
+            joystick,
+            axis_deadzone,
+        ));
+    }
+
+    fn contains(&self, input: GenericInput) -> bool {
+        self.inputs.0.contains(&input)
+    }
+
+    fn consume(self) -> Vec<GenericInput> {
+        self.inputs.0
+    }
+
+    #[must_use]
+    fn insert(&mut self, input: GenericInput) -> CollectionDone {
+        if self.gamepad_starting_states.remove(&input) {
+            return CollectionDone::No;
+        }
+
+        if let Some(opposite) = opposite_input(input) {
+            if self.inputs.0.contains(&opposite) {
+                return CollectionDone::Yes;
+            }
+        }
+
+        self.inputs.insert(input);
+        if self.inputs.len() == jgenesis_native_driver::input::MAX_MAPPING_LEN {
+            CollectionDone::Yes
+        } else {
+            CollectionDone::No
+        }
+    }
+}
+
+fn opposite_input(input: GenericInput) -> Option<GenericInput> {
+    match input {
+        GenericInput::Gamepad { gamepad_idx, action: GamepadAction::Axis(axis_idx, direction) } => {
+            Some(GenericInput::Gamepad {
+                gamepad_idx,
+                action: GamepadAction::Axis(axis_idx, direction.inverse()),
+            })
+        }
+        _ => None,
+    }
+}
+
 fn collect_input(
     event_pump: &mut EventPump,
     joysticks: &mut Joysticks,
     axis_deadzone: i16,
     mut window: Option<InputWindow>,
 ) -> Option<Vec<GenericInput>> {
-    let mut inputs = VecSet::new();
+    let mut inputs = CollectedInputs::new(joysticks, axis_deadzone);
 
     loop {
         for event in event_pump.poll_iter() {
@@ -722,25 +589,34 @@ fn collect_input(
                     return None;
                 }
                 Event::KeyDown { keycode: Some(keycode), .. } => {
-                    inputs.insert(GenericInput::Keyboard(keycode));
+                    if inputs.insert(GenericInput::Keyboard(keycode)) == CollectionDone::Yes {
+                        return Some(inputs.consume());
+                    }
                 }
                 Event::KeyUp { .. } | Event::JoyButtonUp { .. } | Event::MouseButtonUp { .. } => {
-                    return Some(inputs.0);
+                    return Some(inputs.consume());
                 }
                 Event::JoyDeviceAdded { which: device_id, .. } => {
                     if let Err(err) = joysticks.handle_device_added(device_id) {
                         log::error!("Error adding joystick with device id {device_id}: {err}");
                     }
+
+                    if let Some(joystick) = joysticks.device(device_id) {
+                        inputs.add_device(device_id, joystick, axis_deadzone);
+                    }
                 }
                 Event::JoyDeviceRemoved { which: instance_id, .. } => {
                     joysticks.handle_device_removed(instance_id);
                 }
                 Event::JoyButtonDown { which: instance_id, button_idx, .. } => {
                     if let Some(device_id) = joysticks.map_to_device_id(instance_id) {
-                        inputs.insert(GenericInput::Gamepad {
+                        if inputs.insert(GenericInput::Gamepad {
                             gamepad_idx: device_id,
                             action: GamepadAction::Button(button_idx),
-                        });
+                        }) == CollectionDone::Yes
+                        {
+                            return Some(inputs.consume());
+                        }
                     }
                 }
                 Event::JoyAxisMotion { which: instance_id, axis_idx, value, .. } => {
@@ -751,19 +627,22 @@ fn collect_input(
                     let pressed = value.saturating_abs() > axis_deadzone;
                     if pressed {
                         let direction = AxisDirection::from_value(value);
-                        inputs.insert(GenericInput::Gamepad {
+                        if inputs.insert(GenericInput::Gamepad {
                             gamepad_idx,
                             action: GamepadAction::Axis(axis_idx, direction),
-                        });
+                        }) == CollectionDone::Yes
+                        {
+                            return Some(inputs.consume());
+                        }
                     } else if [AxisDirection::Positive, AxisDirection::Negative].into_iter().any(
                         |direction| {
-                            inputs.0.contains(&GenericInput::Gamepad {
+                            inputs.contains(GenericInput::Gamepad {
                                 gamepad_idx,
                                 action: GamepadAction::Axis(axis_idx, direction),
                             })
                         },
                     ) {
-                        return Some(inputs.0);
+                        return Some(inputs.consume());
                     }
                 }
                 Event::JoyHatMotion { which: instance_id, hat_idx, state, .. } => {
@@ -773,35 +652,36 @@ fn collect_input(
 
                     if state == HatState::Centered {
                         if HatDirection::ALL.into_iter().any(|direction| {
-                            inputs.0.contains(&GenericInput::Gamepad {
+                            inputs.contains(GenericInput::Gamepad {
                                 gamepad_idx,
                                 action: GamepadAction::Hat(hat_idx, direction),
                             })
                         }) {
-                            return Some(inputs.0);
+                            return Some(inputs.consume());
                         }
 
                         continue;
                     }
 
                     if let Some(direction) = hat_direction_for(state) {
-                        inputs.insert(GenericInput::Gamepad {
+                        if inputs.insert(GenericInput::Gamepad {
                             gamepad_idx,
                             action: GamepadAction::Hat(hat_idx, direction),
-                        });
+                        }) == CollectionDone::Yes
+                        {
+                            return Some(inputs.consume());
+                        }
                     }
                 }
                 Event::MouseButtonDown { mouse_btn, .. } => {
-                    inputs.insert(GenericInput::Mouse(mouse_btn));
+                    if inputs.insert(GenericInput::Mouse(mouse_btn)) == CollectionDone::Yes {
+                        return Some(inputs.consume());
+                    }
                 }
                 _ => {}
             }
         }
 
-        if inputs.len() == jgenesis_native_driver::input::MAX_MAPPING_LEN {
-            return Some(inputs.0);
-        }
-
         if let Some(window) = &mut window {
             let result = window.update(|ctx| {
                 egui::CentralPanel::default().show(ctx, |ui| {
@@ -828,6 +708,61 @@ fn hat_direction_for(state: HatState) -> Option<HatDirection> {
     }
 }
 
+fn joystick_starting_state(
+    device_id: u32,
+    joystick: &Joystick,
+    axis_deadzone: i16,
+) -> impl Iterator<Item = GenericInput> + use<'_> {
+    buttons_starting_state(device_id, joystick)
+        .chain(axes_starting_state(device_id, joystick, axis_deadzone))
+        .chain(hats_starting_state(device_id, joystick))
+}
+
+fn buttons_starting_state(
+    gamepad_idx: u32,
+    joystick: &Joystick,
+) -> impl Iterator<Item = GenericInput> + use<'_> {
+    (0..joystick.num_buttons()).filter_map(move |button_idx| {
+        let pressed = joystick.button(button_idx).ok()?;
+        pressed.then_some(GenericInput::Gamepad {
+            gamepad_idx,
+            action: GamepadAction::Button(button_idx as u8),
+        })
+    })
+}
+
+fn axes_starting_state(
+    gamepad_idx: u32,
+    joystick: &Joystick,
+    deadzone: i16,
+) -> impl Iterator<Item = GenericInput> + use<'_> {
+    (0..joystick.num_axes()).filter_map(move |axis_idx| {
+        let axis_value = joystick.axis(axis_idx).ok()?;
+        if axis_value.saturating_abs() < deadzone {
+            return None;
+        }
+
+        let direction = AxisDirection::from_value(axis_value);
+        Some(GenericInput::Gamepad {
+            gamepad_idx,
+            action: GamepadAction::Axis(axis_idx as u8, direction),
+        })
+    })
+}
+
+fn hats_starting_state(
+    gamepad_idx: u32,
+    joystick: &Joystick,
+) -> impl Iterator<Item = GenericInput> + use<'_> {
+    (0..joystick.num_hats()).filter_map(move |hat_idx| {
+        let state = joystick.hat(hat_idx).ok()?;
+        hat_direction_for(state).map(|hat_direction| GenericInput::Gamepad {
+            gamepad_idx,
+            action: GamepadAction::Hat(hat_idx as u8, hat_direction),
+        })
+    })
+}
+
 fn render_input_window(joysticks: &Joysticks, ui: &mut egui::Ui) {
     egui::ScrollArea::vertical().show(ui, |ui| {
         ui.label(
