@@ -3,7 +3,7 @@ use gb_core::inputs::GameBoyButton;
 use genesis_core::GenesisControllerType;
 use genesis_core::input::GenesisButton;
 use jgenesis_common::input::Player;
-use jgenesis_proc_macros::{ConfigDisplay, EnumDisplay, EnumFromStr};
+use jgenesis_proc_macros::{ConfigDisplay, EnumAll, EnumDisplay};
 use nes_core::input::NesButton;
 use sdl2::keyboard::Keycode;
 use sdl2::mouse::MouseButton;
@@ -388,8 +388,9 @@ impl NesInputMapping {
 }
 
 #[derive(
-    Debug, Clone, Copy, PartialEq, Eq, Default, Serialize, Deserialize, EnumDisplay, EnumFromStr,
+    Debug, Clone, Copy, PartialEq, Eq, Default, Serialize, Deserialize, EnumDisplay, EnumAll,
 )]
+#[cfg_attr(feature = "clap", derive(jgenesis_proc_macros::CustomValueEnum))]
 pub enum NesControllerType {
     #[default]
     Gamepad,
@@ -524,8 +525,9 @@ impl SnesInputMapping {
 }
 
 #[derive(
-    Debug, Clone, Copy, PartialEq, Eq, Default, Serialize, Deserialize, EnumDisplay, EnumFromStr,
+    Debug, Clone, Copy, PartialEq, Eq, Default, Serialize, Deserialize, EnumDisplay, EnumAll,
 )]
+#[cfg_attr(feature = "clap", derive(jgenesis_proc_macros::CustomValueEnum))]
 pub enum SnesControllerType {
     #[default]
     Gamepad,
@@ -643,7 +645,14 @@ macro_rules! define_hotkey_mapping {
         None
     };
     (@default $default:ident) => {
-        key_input!($default)
+        Some(vec![GenericInput::Keyboard(Keycode::$default)])
+    };
+    (@default ($($default:ident $(+)?)*)) => {
+        Some(vec![
+            $(
+                GenericInput::Keyboard(Keycode::$default),
+            )*
+        ])
     };
     ($($value:ident: $hotkey:ident default $default:tt,)* $(,)?) => {
         #[derive(Debug, Clone, PartialEq, Eq, Default, Serialize, Deserialize)]
@@ -685,7 +694,8 @@ macro_rules! define_hotkey_mapping {
 }
 
 define_hotkey_mapping!(
-    quit: Quit default Escape,
+    power_off: PowerOff default Escape,
+    exit: Exit default (LCtrl + Q),
     toggle_fullscreen: ToggleFullscreen default F9,
     save_state: SaveState default F5,
     load_state: LoadState default F6,
