@@ -134,6 +134,7 @@ struct HotkeyState<Emulator> {
     fast_forward_multiplier: u64,
     rewinder: Rewinder<Emulator>,
     debugger_window: Option<DebuggerWindow<Emulator>>,
+    window_scale_factor: Option<f32>,
     debug_render_fn: fn() -> Box<DebugRenderFn<Emulator>>,
 }
 
@@ -176,6 +177,7 @@ impl<Emulator: PartialClone> HotkeyState<Emulator> {
                 common_config.rewind_buffer_length_seconds,
             )),
             debugger_window: None,
+            window_scale_factor: common_config.window_scale_factor,
             debug_render_fn,
         })
     }
@@ -183,7 +185,7 @@ impl<Emulator: PartialClone> HotkeyState<Emulator> {
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum NativeTickEffect {
-    None,
+    PowerOff,
     Exit,
 }
 
@@ -262,17 +264,18 @@ impl<Inputs, Button, Config, Emulator: EmulatorTrait>
 }
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
-enum HotkeyResult {
-    None,
-    Quit,
+enum HotkeyEffect {
+    PowerOff,
+    Exit,
 }
 
 fn open_debugger_window<Emulator>(
     video: &VideoSubsystem,
+    scale_factor: Option<f32>,
     debug_render_fn: fn() -> Box<DebugRenderFn<Emulator>>,
 ) -> Option<DebuggerWindow<Emulator>> {
     let render_fn = debug_render_fn();
-    match DebuggerWindow::new(video, render_fn) {
+    match DebuggerWindow::new(video, scale_factor, render_fn) {
         Ok(debugger_window) => Some(debugger_window),
         Err(err) => {
             log::error!("Error opening debugger window: {err}");
@@ -399,12 +402,16 @@ where
     ) -> NativeEmulatorResult<Self> {
         let (sdl, video, audio, joystick, event_pump) = init_sdl(&common_config)?;
 
-        let window_size = common_config.window_size.unwrap_or(default_window_size);
+        let mut initial_window_size = common_config.window_size.unwrap_or(default_window_size);
+        if let Some(scale_factor) = common_config.window_scale_factor {
+            initial_window_size = initial_window_size.scale(scale_factor);
+        }
+
         let window = create_window(
             &video,
             window_title,
-            window_size.width,
-            window_size.height,
+            initial_window_size.width,
+            initial_window_size.height,
             common_config.launch_in_fullscreen.then_some(common_config.fullscreen_mode),
         )?;
 
@@ -462,122 +469,111 @@ where
     ///
     /// This method will propagate any errors encountered when rendering frames, pushing audio
     /// samples, or writing save files.
-    pub fn render_frame(&mut self) -> NativeEmulatorResult<NativeTickEffect> {
-        loop {
-            let rewinding = self.hotkey_state.rewinder.is_rewinding();
-            let should_tick_emulator =
-                !rewinding && (!self.hotkey_state.paused || self.hotkey_state.should_step_frame);
+    pub fn render_frame(&mut self) -> NativeEmulatorResult<Option<NativeTickEffect>> {
+        let rewinding = self.hotkey_state.rewinder.is_rewinding();
+        let should_run_emulator =
+            !rewinding && (!self.hotkey_state.paused || self.hotkey_state.should_step_frame);
 
-            let frame_rendered = should_tick_emulator
-                && self
-                    .emulator
-                    .tick(
-                        &mut self.renderer,
-                        &mut self.audio_output,
-                        self.input_mapper.inputs(),
-                        &mut self.save_writer,
-                    )
-                    .map_err(|err| NativeEmulatorError::Emulator(err.into()))?
-                    == TickEffect::FrameRendered;
+        if should_run_emulator {
+            while self
+                .emulator
+                .tick(
+                    &mut self.renderer,
+                    &mut self.audio_output,
+                    self.input_mapper.inputs(),
+                    &mut self.save_writer,
+                )
+                .map_err(|err| NativeEmulatorError::Emulator(err.into()))?
+                != TickEffect::FrameRendered
+            {}
 
-            if frame_rendered {
-                self.fps_tracker.record_frame();
+            self.fps_tracker.record_frame();
+            self.hotkey_state.rewinder.record_frame(&self.emulator);
 
-                self.audio_output.adjust_dynamic_resampling_ratio();
-                self.emulator.update_audio_output_frequency(self.audio_output.output_frequency());
+            self.audio_output.adjust_dynamic_resampling_ratio();
+            self.emulator.update_audio_output_frequency(self.audio_output.output_frequency());
+        }
+
+        self.hotkey_state.should_step_frame = false;
+
+        if let Some(debugger_window) = &mut self.hotkey_state.debugger_window {
+            if let Err(err) = debugger_window.update(&mut self.emulator) {
+                log::error!("Debugger window error: {err}");
             }
+        }
 
-            if should_tick_emulator && !frame_rendered {
-                continue;
-            }
+        // Gymnastics to avoid borrow checker errors that would otherwise occur due to
+        // calling `&mut self` methods while mutably borrowing the event pump
+        let event_buffer_ref = Rc::clone(&self.event_buffer);
+        let mut event_buffer = event_buffer_ref.borrow_mut();
+        event_buffer.extend(self.event_pump.poll_iter());
 
-            self.hotkey_state.should_step_frame = false;
+        for event in event_buffer.drain(..) {
+            self.input_mapper.handle_event(
+                &event,
+                self.renderer.window_id(),
+                self.renderer.current_display_info(),
+            );
 
             if let Some(debugger_window) = &mut self.hotkey_state.debugger_window {
-                if let Err(err) = debugger_window.update(&mut self.emulator) {
-                    log::error!("Debugger window error: {err}");
-                }
+                debugger_window.handle_sdl_event(&event);
             }
 
-            // Gymnastics to avoid borrow checker errors that would otherwise occur due to
-            // calling `&mut self` methods while mutably borrowing the event pump
-            let event_buffer_ref = Rc::clone(&self.event_buffer);
-            let mut event_buffer = event_buffer_ref.borrow_mut();
-            event_buffer.extend(self.event_pump.poll_iter());
-
-            for event in event_buffer.drain(..) {
-                self.input_mapper.handle_event(
-                    &event,
-                    self.renderer.window_id(),
-                    self.renderer.current_display_info(),
-                );
-
-                if let Some(debugger_window) = &mut self.hotkey_state.debugger_window {
-                    debugger_window.handle_sdl_event(&event);
+            match event {
+                Event::Quit { .. } => {
+                    return Ok(Some(NativeTickEffect::PowerOff));
                 }
-
-                // if self.handle_hotkeys(&event)? == HotkeyResult::Quit {
-                //     return Ok(NativeTickEffect::Exit);
-                // }
-
-                match event {
-                    Event::Quit { .. } => {
-                        return Ok(NativeTickEffect::Exit);
-                    }
-                    Event::Window { win_event, window_id, .. } => {
-                        if win_event == WindowEvent::Close {
-                            if window_id == self.renderer.window_id() {
-                                return Ok(NativeTickEffect::Exit);
-                            }
-
-                            if self
-                                .hotkey_state
-                                .debugger_window
-                                .as_ref()
-                                .is_some_and(|debugger| window_id == debugger.window_id())
-                            {
-                                self.hotkey_state.debugger_window = None;
-                            }
-                        }
-
+                Event::Window { win_event, window_id, .. } => {
+                    if win_event == WindowEvent::Close {
                         if window_id == self.renderer.window_id() {
-                            handle_window_event(win_event, &mut self.renderer);
+                            return Ok(Some(NativeTickEffect::PowerOff));
+                        }
+
+                        if self
+                            .hotkey_state
+                            .debugger_window
+                            .as_ref()
+                            .is_some_and(|debugger| window_id == debugger.window_id())
+                        {
+                            self.hotkey_state.debugger_window = None;
                         }
                     }
-                    _ => {}
-                }
-            }
 
-            let hotkey_events = self.input_mapper.hotkey_events();
-            {
-                let mut hotkey_events = hotkey_events.borrow_mut();
-                for &hotkey_event in &*hotkey_events {
-                    if self.handle_hotkey_event(hotkey_event)? == HotkeyResult::Quit {
-                        return Ok(NativeTickEffect::Exit);
+                    if window_id == self.renderer.window_id() {
+                        handle_window_event(win_event, &mut self.renderer);
                     }
                 }
-                hotkey_events.clear();
+                _ => {}
             }
-
-            if frame_rendered {
-                self.hotkey_state.rewinder.record_frame(&self.emulator);
-            }
-
-            if rewinding {
-                self.hotkey_state.rewinder.tick(
-                    &mut self.emulator,
-                    &mut self.renderer,
-                    &self.config,
-                )?;
-            }
-
-            if !should_tick_emulator {
-                // Don't spin loop when the emulator is not actively running
-                jgenesis_common::sleep(Duration::from_millis(1));
-            }
-
-            return Ok(NativeTickEffect::None);
         }
+
+        let hotkey_events = self.input_mapper.hotkey_events();
+        {
+            let mut hotkey_events = hotkey_events.borrow_mut();
+            for &hotkey_event in &*hotkey_events {
+                match self.handle_hotkey_event(hotkey_event)? {
+                    Some(HotkeyEffect::PowerOff) => return Ok(Some(NativeTickEffect::PowerOff)),
+                    Some(HotkeyEffect::Exit) => return Ok(Some(NativeTickEffect::Exit)),
+                    None => {}
+                }
+            }
+            hotkey_events.clear();
+        }
+
+        if rewinding {
+            self.hotkey_state.rewinder.tick(
+                &mut self.emulator,
+                &mut self.renderer,
+                &self.config,
+            )?;
+        }
+
+        if !should_run_emulator {
+            // Don't spin loop when the emulator is paused or rewinding
+            jgenesis_common::sleep(Duration::from_millis(1));
+        }
+
+        Ok(None)
     }
 
     pub fn soft_reset(&mut self) {
@@ -590,8 +586,11 @@ where
 
     pub fn open_memory_viewer(&mut self) {
         if self.hotkey_state.debugger_window.is_none() {
-            self.hotkey_state.debugger_window =
-                open_debugger_window(&self.video, self.hotkey_state.debug_render_fn);
+            self.hotkey_state.debugger_window = open_debugger_window(
+                &self.video,
+                self.hotkey_state.window_scale_factor,
+                self.hotkey_state.debug_render_fn,
+            );
         }
     }
 
@@ -662,11 +661,14 @@ where
         &self.hotkey_state.save_state_metadata
     }
 
-    fn handle_hotkey_event(&mut self, event: HotkeyEvent) -> NativeEmulatorResult<HotkeyResult> {
+    fn handle_hotkey_event(
+        &mut self,
+        event: HotkeyEvent,
+    ) -> NativeEmulatorResult<Option<HotkeyEffect>> {
         match event {
             HotkeyEvent::Pressed(hotkey) => {
-                if self.handle_hotkey_pressed(hotkey)? == HotkeyResult::Quit {
-                    return Ok(HotkeyResult::Quit);
+                if let Some(effect) = self.handle_hotkey_pressed(hotkey)? {
+                    return Ok(Some(effect));
                 }
             }
             HotkeyEvent::Released(hotkey) => match hotkey {
@@ -681,13 +683,17 @@ where
             },
         }
 
-        Ok(HotkeyResult::None)
+        Ok(None)
     }
 
-    fn handle_hotkey_pressed(&mut self, hotkey: Hotkey) -> NativeEmulatorResult<HotkeyResult> {
+    fn handle_hotkey_pressed(
+        &mut self,
+        hotkey: Hotkey,
+    ) -> NativeEmulatorResult<Option<HotkeyEffect>> {
         let hotkey = hotkey.to_compact();
         match hotkey {
-            CompactHotkey::Quit => return Ok(HotkeyResult::Quit),
+            CompactHotkey::PowerOff => return Ok(Some(HotkeyEffect::PowerOff)),
+            CompactHotkey::Exit => return Ok(Some(HotkeyEffect::Exit)),
             CompactHotkey::ToggleFullscreen => {
                 self.renderer
                     .toggle_fullscreen(self.hotkey_state.fullscreen_mode)
@@ -772,7 +778,11 @@ where
             CompactHotkey::OpenDebugger => {
                 if self.hotkey_state.debugger_window.is_none() {
                     let debug_render_fn = (self.hotkey_state.debug_render_fn)();
-                    match DebuggerWindow::new(&self.video, debug_render_fn) {
+                    match DebuggerWindow::new(
+                        &self.video,
+                        self.hotkey_state.window_scale_factor,
+                        debug_render_fn,
+                    ) {
                         Ok(debugger_window) => {
                             self.hotkey_state.debugger_window = Some(debugger_window);
                         }
@@ -784,7 +794,7 @@ where
             }
         }
 
-        Ok(HotkeyResult::None)
+        Ok(None)
     }
 }
 
